(* 
 * Copyright (c) 2001 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: gontbind.ml,v 1.13 2002/02/08 07:31:15 malekith Exp $
 *
 * Module implementation: Gontbind
 * Description:
 * The purpouse of this file to make other part of compiler 
 * symbol-table-free (however type variables are handled in
 * Gonttypes, this module doesn't touch it). It generally changes
 * symbol options from None to Some sym. We also set variable
 * stating how deep in closure stack should we go to finally
 * get to the variable. We also mark variables to be put into
 * closures here.
 *)

open Gontutil

(*
 * Level stuff:
 *   fnc_level is level of function nesting for symbol.
 *     -2 for undefined symbols
 *     -1 for symbols from other modules
 *      0 for global symbols (in current module)
 *      1 for variables and parmaters to 1st level functions
 *      2 ditto for 2nd level...
 *   scope_level is level of scope nesting of symbol, levels -2 up to
 *   0 are the same as for fnc_level, next define level of block nesting.
 *
 * Example            // (fnc_level, scope_level)
 * int foo;           // (0, 0)
 * int bar(           // (0, 0)
 *     int x) {       // (1, 1)
 *     int baz;       // (1, 1)
 *     int f() {      // (1, 1)
 *         int k;     // (2, 2)
 *     }
 *     { int z; }     // (1, 2)
 *  }
 *)

type ctx = {ctx_types: typedef list;
            ctx_syms: symbol list;
            ctx_fnc_level: int;
            ctx_scope_level: int;
            ctx_opened_modules: string list;
            ctx_module: string}

let make_symbol c t n =
  new_symbol (c.ctx_scope_level, c.ctx_fnc_level, t, n)
;;

let init_bind modname = {ctx_syms = []; 
                         ctx_types = [];
                         ctx_fnc_level = 0; 
                         ctx_scope_level = 0;
                         ctx_opened_modules = [];
                         ctx_module = modname}

let find_mod_sym ctx m n =
   let sym = Gontmodule.import_symbol m n in
   if ctx.ctx_module <> m then
     sym#set_property IsNonLocal;
   sym

let find_mod_type ctx m n =
     ignore(ctx);
     Gontmodule.import_type_symbol m n
;;

let find_sym ctx n =
  let qn = (ctx.ctx_module ^ "::" ^ n) in
  try
    List.find (fun sym -> sym#get_name = n || sym#get_name = qn) ctx.ctx_syms
  with Not_found -> 
    let m =
      List.find (fun m -> 
          try
            ignore(find_mod_sym ctx m n);
            true
          with Not_found -> false) ctx.ctx_opened_modules in
      find_mod_sym ctx m n

let find_type ctx m n =
  match m with
      "" -> 
        (try
          List.find (fun {td_name = x} -> n = x) ctx.ctx_types
         with Not_found ->
           let m = 
             List.find (fun m -> 
               try 
                 ignore(find_mod_type ctx m n); 
                 true 
               with Not_found -> false) ctx.ctx_opened_modules in
           find_mod_type ctx m n)
    | _ -> find_mod_type ctx m n
;;

let prepare_ctx loc (c, parms) =
  let rec helper (c, parms) =
    match parms with
        ({parm_type=t; parm_name=n} as x)::xs -> 
          (let (fnc_lev, scope_lev) =
            try
              let sym = find_sym c n in
                (sym#get_fnc_level, sym#get_scope_level)
            with Not_found -> (-2, -2) in
          if fnc_lev == c.ctx_fnc_level then
            err (loc, xf "parameter `%s' defined twice" n);
          if scope_lev >= 0 then
            (* Note, that we don't warn about shodowing symbols
             * from other modules, that has been opened. *)
            warn (loc, xf "`%s' shadows other symbol" n));
          let sym = make_symbol c t n in
          sym#set_property IsInitialized;
          x.parm_symbol <- Some sym;
          helper ({c with ctx_syms = sym :: c.ctx_syms }, xs)
      | [] -> c in
  helper ({c with ctx_fnc_level   = c.ctx_fnc_level + 1;
                  ctx_scope_level = c.ctx_scope_level + 1}, 
          parms)
;;

(* This does inplace modification of s. Updated context is returned. *)
let rec bind_stmt (c, s) =
  match s.raw_stmt with
      Skip -> c
    | Exp(e) -> bind_exp (c, e)
    | VarDef v ->
          (match v.var_symbol with
              None -> ()
            | Some _ -> assert false);
          let (c, t) = bind_type (c, v.var_type) in
          (let scope_lev =
            try
              let sym = find_sym c v.var_name in
                sym#get_scope_level
            with Not_found -> -2 in
          if scope_lev == c.ctx_scope_level then
            err (s.stmt_loc, xf "variable `%s' defined twice" v.var_name);
          if scope_lev >= 0 then
            warn (s.stmt_loc, xf "`%s' shadows other symbol" v.var_name));
          let sym = make_symbol c t v.var_name in
          v.var_symbol <- Some sym;
          {c with ctx_syms = sym :: c.ctx_syms}
    | Return(e) -> bind_exp (c, e)
    | ReturnVoid -> c
    | Block(ss) ->
        let rec bind_block (c, ss) =
          match ss with
              x::xs -> 
                let c' = bind_stmt (c, x) in
                bind_block (c', xs)
            | [] -> c in
        let new_scope = {c with ctx_scope_level = c.ctx_scope_level + 1} in
        let c' = bind_block(new_scope, ss) in
        ignore(c'); 
        c
    | Pair(s1, s2) ->
        bind_stmt ((bind_stmt (c, s1)), s2)
    | Switch(expr, patlist) ->
        let rec bind_patterns (c, patl) =
          match patl with
              x::xs ->
                let {cc_pattern = pat; cc_stmt = st} = x in
                let pat_scope = 
                  {c with ctx_scope_level = c.ctx_scope_level + 1} in
                let c' = bind_pattern (pat_scope, pat) in
                let c'' = bind_stmt (c', st) in
                ignore(c'');
                bind_patterns (c, xs)
            | [] -> c in
        let exprc = bind_exp (c, expr) in
        let patc = bind_patterns (exprc, patlist) in
        ignore(patc);
        exprc                   
    | Let (patel, stmt) ->
        let rec expr_binder (c, pel) =
           match pel with
           | x::xs ->
              let (_, ex) = x in
              let cn = bind_exp (c, ex) in
              expr_binder (cn, xs)
           | [] -> c in
        let rec pate_binder (c, pel) =
           match pel with
           | x::xs ->
              let (pa, _) = x in
              let cn = bind_let_pattern (c, pa) in
              expr_binder (cn, xs)
           | [] -> c in
        let c1 = expr_binder (c, patel) in
        let newscope =
          {c1 with ctx_scope_level = c1.ctx_scope_level + 1} in
        let c2 = pate_binder (newscope, patel) in
        let c3 = bind_stmt (c2, stmt) in
        ignore(c3);
        c1
    | BreakIfNot e ->
        bind_exp (c, e)
    | Loop s ->
        bind_stmt (c, s)
    | If(e, s1, s2) ->
        bind_stmt (bind_stmt (bind_exp (c, e), s1), s2)
    | Open m ->
        Gontmodule.open_module m;
        {c with ctx_opened_modules = m :: c.ctx_opened_modules}
and bind_exp (c, e) =
  match e.raw_exp with
      SimpleBinOp(op, e1, e2) -> 
        let c' = bind_exp (c, e1) in
        bind_exp (c', e2)
    | SimpleUnOp(op, e1) ->
        bind_exp (c, e1)
    | Ref(r) ->
        (match r.ref_symbol with
            None ->
              let sym = 
                try
                  find_sym c r.ref_name
                with Not_found ->
                  err (e.exp_loc, xf "unbound symbol `%s'" r.ref_name) in
              r.ref_symbol <- Some sym;
              let lev = c.ctx_fnc_level - sym#get_fnc_level in
                assert (lev >= 0);
                r.ref_level <- if sym#get_fnc_level == 0 then 0 else lev
          (* We have already been here! *)
          | Some _ -> assert false); 
          (* Context is not updated in any way. *)
          c
    | CrossModuleRef(r) ->
        (match r.xmr_symbol with
            None ->
              (try
                 let sym = find_mod_sym c r.xmr_module r.xmr_name in
                 r.xmr_symbol <- Some sym
               with Not_found ->
                 err (e.exp_loc, xf "there is no symbol %s in module %s"
                                    r.xmr_name r.xmr_module))
          | Some _ -> assert false);
          c
    | Call(e1, es) -> 
        let c' = bind_exp (c, e1) in
        let rec bind_parms (c, pp) =
            match pp with
              p::ps -> 
                let c' = bind_exp (c, p) in
                bind_parms (c', ps)
           | [] -> c in
        bind_parms (c', es)
    | IntConst(_) -> c
    | BoolConst(_) -> c
    | FloatConst(_) -> c
    | StringConst(_) -> c
    | VoidConst -> c
    | UnionExp (un) -> 
        let _ =
            match un.ur_sym with
            | None ->
                (* find out if there exists union containing id field *) 	    
                let sym = 
                  try
                    if un.ur_module = "" then
                      find_sym c un.ur_name
                    else
                      find_mod_sym c un.ur_module un.ur_name
                  with Not_found -> 
                   err (e.exp_loc, xf "there is no union field %s::%s"
                                      un.ur_module un.ur_name) in
                 un.ur_sym <- Some sym
            | Some _ -> assert false in
        bind_exp(c, un.ur_exp)
    | TupleExp (tup) ->
        let pair (ct, ex) = (bind_exp(ct,ex), ex) in
        let (c', _) = list_apply pair (c, tup) in
        c'
    | Lambda {lambda_parms=p; lambda_ret_type=r; lambda_body=s} ->
        let new_ctx = prepare_ctx e.exp_loc (c, p) in
        let (c', t) = bind_type (new_ctx, make_fn_type (r, p)) in
        let c'' = bind_stmt (c', s) in
        ignore(c'');
       (* We return the same contex. Local function cannot affect
        * parent enviroment. *)
       c
and add_fields (c, td) =
  match td.td_type with
    Union uni ->       (* executed during union type declaration *)
      let rec bind_union_fields (no, c, fil) =  
        match fil with
          x::xs ->
            let (t, id) = x in
            (let scope_lev =
              try
                let sym = find_sym c id in
                sym#get_scope_level
              with Not_found -> -2 in
            if scope_lev == c.ctx_scope_level then
              err (td.td_loc, xf "field `%s' already defined" id);
            if scope_lev >= 0 then
              warn (td.td_loc, xf "field `%s' shadows other symbol" id));
            let sym = make_symbol c t id in
            sym#make_union_field (no, td);
            let c' = {c with ctx_syms = sym :: c.ctx_syms} in
            bind_union_fields(no + 1, c', xs)
        | [] -> c in
      bind_union_fields (1, c, uni)
    | _ -> c
and bind_type (c, t) =
  let f t =
    match t with
        NamedType ({nt_tydef = None} as r) ->
          (try
             r.nt_tydef <- Some (find_type c r.nt_module r.nt_name);
             NamedType r
           with Not_found ->
             if r.nt_module = "" then
               err (r.nt_loc, xf "use of unknown type %s" r.nt_name)
             else
               err (r.nt_loc, xf "module %s does not define type %s"
                                 r.nt_module r.nt_name))
        (* We used to have assert false here (as we already bound this
         * type), however lambda expressions generated by parsers
         * are somewhat tricky -- they use the same type for type of
         * functional variable being defined and for parameters of
         * functional value, so bind_type() finishes up called two times
         * on the same type. This should be probably fixed in parser,
         * not here, but I'm not sure. --mal *)
      | NamedType _ -> t
      | Union l -> 
          Union (List.map (fun (t, n) -> (t, c.ctx_module ^ "::" ^ n)) l)
      | Int | Void | Bool | Float | String | TyVar _ | AbstractType
      | Tuple _ | Func _ | Struct _ | OptStruct _ -> t in
  (c, type_walk f t)
and bind_pattern (c, p) =
  match p.raw_pattern with
    VoidPattern -> 
      c
  | VarPattern vap ->
      (match vap.vp_symbol with
        None ->
          let sym = make_symbol c AbstractType vap.vp_name in
          sym#set_property IsInitialized;
          vap.vp_symbol <- Some sym;
          {c with ctx_syms = sym :: c.ctx_syms}  
      (* We have already been here! *)
      | Some _ -> assert false
      )
  | ExprPattern e ->
      (match e.raw_exp with
      | Ref r ->
         (try
            ignore(find_sym c r.ref_name);
            bind_exp(c, e)
          with Not_found ->
            p.raw_pattern <- VarPattern {vp_name = r.ref_name;
                                         vp_symbol = None};
            bind_pattern(c, p))
       | _ -> bind_exp (c, e))
  | TuplePattern tup ->
      let pair (ct, pa) = (bind_pattern (ct, pa), pa) in
      let (c', _) = list_apply pair (c, tup.tpr_patterns) in
      c'
  | UnionPattern un ->
      let sym = 
        try
          find_sym c un.upr_name 
        with Not_found -> 
          (*List.iter(fun s -> print_endline s#get_name) c.ctx_syms; *)
          err(p.pat_loc, xf "union field `%s' not defined" un.upr_name) in
      un.upr_sym <- Some sym;
      bind_pattern (c, un.upr_pattern)

and bind_let_pattern (c, p) =
  match p.raw_pattern with
    VoidPattern ->
      c
  | VarPattern vap ->
      (match vap.vp_symbol with
        None ->
          (let scope_lev =
            try
              let sym = find_sym c vap.vp_name in
              sym#get_scope_level
            with Not_found -> -2 in
          if scope_lev == c.ctx_scope_level then
            err (p.pat_loc, xf "pattern variable `%s' defined twice" 
                   vap.vp_name);
          if scope_lev >= 0 then
            warn (p.pat_loc, xf "`%s' in pattern shadows other symbol" 
                    vap.vp_name));
          let sym = make_symbol c AbstractType vap.vp_name in
          vap.vp_symbol <- Some sym;
          sym#set_property IsInitialized;
          {c with ctx_syms = sym :: c.ctx_syms}  
      (* We have already been here! *)
      | Some _ -> assert false
      )
  | ExprPattern e ->
      (match e.raw_exp with
      | Ref r ->
          p.raw_pattern <- VarPattern {vp_name = r.ref_name;
                                       vp_symbol = None};
          p.pat_loc <- e.exp_loc;
          bind_let_pattern(c, p)
       | _ -> 
         err (e.exp_loc, "no expressions allowed in let pattern")
      )
  | TuplePattern tup ->
      let pair (ct, pa) = (bind_let_pattern (ct, pa), pa) in
      let (c', _) = list_apply pair (c, tup.tpr_patterns) in
      c'
  | UnionPattern un ->
      err (p.pat_loc, "no unions allowed in let pattern")
      (* altough there could be union with one field, but 
         I think we don't have to consider it *)  
;;

(* Modify stmts in place, adding symbol entries to Ref terms. This musn't be
 * called more then once in a toplevel function. *)
let bind c decl =
  let add_global loc t n =
    try
      ignore(find_sym c n);
      err (loc, xf "redefintion of global symbol `%s'" n)
    with Not_found -> make_symbol c t n in
  match decl with
      FuncDecl f ->
        let fntype = make_fn_type (f.fn_ret_type, f.fn_parms) in
        let (c, fntype) = bind_type (c, fntype) in
        let sym = add_global f.fn_loc
                             fntype
                             (c.ctx_module ^ "::" ^ f.fn_name) in
        let c' = {c with ctx_syms = sym :: c.ctx_syms} in
        f.fn_symbol <- Some sym;
        sym#set_property IsNonLvalue;
        let initial_ctx = prepare_ctx f.fn_loc (c', f.fn_parms) in
        ignore (bind_stmt (initial_ctx, f.fn_body));
        c'
    | GlobalDecl g ->
        let (c, t) = bind_type (c, g.global_type) in
        let sym = add_global g.global_loc 
                             t 
                             (c.ctx_module ^ "::" ^ g.global_name) in
        g.global_symbol <- Some sym;
        sym#set_property IsGlobalVar;
        {c with ctx_syms = sym :: c.ctx_syms}
    | ExternC x ->
        let (c, t) = bind_type (c, x.xc_type) in
        let sym = add_global x.xc_loc t x.xc_name in
        x.xc_symbol <- Some sym;
        {c with ctx_syms = sym :: c.ctx_syms}
    | TopLevelOpen m ->
        Gontmodule.open_module m;
        {c with ctx_opened_modules = m :: c.ctx_opened_modules}
    | TypeDef ({td_name = n} as t) ->
        if t.td_module = "" then
          t.td_module <- c.ctx_module;
        (try 
          let s = find_type c "" n in
          if s.td_module = c.ctx_module then
            err (t.td_loc, xf "redefintion of type `%s'" n)
          else
            warn (t.td_loc, xf "type `%s' shadows `%s::%s'" 
                               n s.td_module s.td_name)
         with Not_found -> ());
        (* Allow recursive types. *)
        let c = {c with ctx_types = t :: c.ctx_types} in
        let (c, tt) = bind_type (c, t.td_type) in
        t.td_type <- tt;
        add_fields (c, t)
    | FuncProto fp ->
        let (c, t) = bind_type (c, fp.fp_type) in
        let sym = add_global fp.fp_loc t
                             (c.ctx_module ^ "::" ^ fp.fp_name) in
        fp.fp_symbol <- Some sym;
        {c with ctx_syms = sym :: c.ctx_syms} 
;;

let finalize_module out_ch c =
  Gontmodule.save_module out_ch c.ctx_module c.ctx_syms c.ctx_types
;;

(* Apply given function, of (exp -> 'a list) type, for all
 * expressions in a tree, returning flat list of results. *)
let rec for_all_exp exp_fnc s =
  let stmt_fnc s = for_all_exp exp_fnc s in
  match s.raw_stmt with
      Skip -> []
    | Exp(e) -> exp_fnc e
    | Return(e) -> exp_fnc e
    | ReturnVoid -> []
    | Block(l) -> (List.flatten (List.map stmt_fnc l))
    | Pair(a, b) -> (stmt_fnc a) @ (stmt_fnc b)
    | Switch(e, ptl) -> 
        let rec patt_divisor pl =
          match pl with
            x::xs -> 
              let {cc_stmt = st} = x in
              (stmt_fnc st) @ patt_divisor xs
          | [] -> [] in
        (exp_fnc e) @ (patt_divisor ptl)
    | Let (pel, st) ->
        let rec pel_divisor pl =
          match pl with
            x::xs -> 
              let (_, e) = x in
              (exp_fnc e) @ pel_divisor xs
          | [] -> [] in
         (pel_divisor pel) @ (stmt_fnc st)        
    | VarDef(_) -> []
    | BreakIfNot e -> exp_fnc e
    | Loop s -> stmt_fnc s
    | If(e, s1, s2) -> exp_fnc e @ stmt_fnc s1 @ stmt_fnc s2
    | Open _ -> []
;;

(* Return list of symbols, at given scoping level, that needs to put into the
 * closure because they are referenced in stmts. *)
let closure lev s =
  let need_closure = ref false in
  let rec exp_closure lambda_lev e =
    let ec = exp_closure lambda_lev in
    match e.raw_exp with
        SimpleBinOp(op, e1, e2) -> (ec e1) @ (ec e2)
      | SimpleUnOp(op, e1) -> ec e1
      | Ref(r) ->
          (match r.ref_symbol with
              None -> assert false
            | Some sym ->
                if r.ref_level > 0 then
                  (if lambda_lev > 0 then need_closure := true;
                   if sym#get_fnc_level = lev && 
                      not (sym#has_property IsInClosure) then
                     (if sym#has_property IsInPattern then
                        err(e.exp_loc, "cannot use pattern variable in local \
                            function");
                      sym#set_property IsInClosure; [sym])
                   else
                     [])
                else
                  [])
      | Call(e1, es) ->
          (ec e1) @ (List.flatten (List.map ec es))
      | CrossModuleRef _ -> []
      | IntConst(_) -> []
      | BoolConst(_) -> []
      | FloatConst(_) -> []
      | StringConst(_) -> []
      | VoidConst -> []
      | TupleExp (tup) ->
          let rec tup_apply tu =
            match tu with
              x::xs -> (ec x) @ (tup_apply xs)
            | [] -> [] in
          tup_apply tup
      | UnionExp (un) -> ec un.ur_exp
      | Lambda {lambda_body=s} ->
         for_all_exp (exp_closure (lambda_lev + 1)) s in
  (!need_closure, for_all_exp (exp_closure 0) s)
;;

