(* 
 * Copyright (c) 2001 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: gonttypes.ml,v 1.14 2002/02/12 11:51:45 malekith Exp $
 *
 * Module implementation: Gonttypes
 * Description: The typechecker. Term unification is used to check whatever
 * given two types are compatibile. If we use any kind of subtyping this
 * will gonna have to be directed somewhat...
 *)

open Gontutil
open Gontbind

(* Pattern ranges for checking pattern matching exchaustive *)
type patrange = 
    AlphaRange                                    (* Match all patterns *)
  | TupleRange of patrange list list
  | UnionRange of (symbol list) * (patrange list)
  | IntRange
  | StringRange
  | FloatRange
  | BoolRange of bool
;;

exception Cannot_join

exception Type_error of string

let sorry s = raise (Type_error s)
;;

let rec string_of_type_list tt =
  String.concat ", " (List.map string_of_type tt)
and string_of_type_name_list l =
  (String.concat "" 
        (List.map (fun (t, n) -> (string_of_type t) ^ " " ^ n ^ "; ") l))
and string_of_fv l =
  match l with
      [] -> ""
    | _ -> "<" ^ (String.concat ", " (List.map (fun s -> "'" ^ s) l)) ^ ">" 
and string_of_type t =
  match t with
      Void -> "void"
    | Int -> "int"
    | Float -> "float"
    | Bool -> "bool"
    | String -> "string"
    | Func f -> 
        (string_of_fv f.ft_tyvars) ^ 
        "*(" ^ (string_of_type_list f.ft_parms) ^ ") -> " ^ 
        (string_of_type f.ft_ret_type)
    | TyVar x -> ("'" ^ x)
    | Tuple l ->
        "*[" ^ (string_of_type_list l) ^ "]"
    | Struct l ->
        "struct { " ^ (string_of_type_name_list l) ^ "}"
    | OptStruct l ->
        "opt_struct { " ^ (string_of_type_name_list l) ^ "}"
    | Union l ->
        "union { " ^ (string_of_type_name_list l) ^ "}"
    | AbstractType -> "abstract_type"
    | NamedType n ->
        (match n.nt_parms with
            [] -> ""
          | l -> "<" ^ (String.concat ", " (List.map string_of_type l)) 
                     ^ ">") ^
        (if n.nt_module = "" then
           n.nt_name
         else
           n.nt_module ^ "::" ^ n.nt_name) 
;;

let rec string_of_symbol sym =
  sym#get_name
and string_of_stmt ind s =
  let make_ind ind =  String.make (ind * 2) ' ' in
  let list_map f l = String.concat "" (List.map (f (ind + 1)) l) in
  let str = 
    match s.raw_stmt with
    | Skip -> 
        "skip"
    | Exp e -> 
        string_of_exp e
    | Return e -> 
        "return " ^ (string_of_exp e)
    | ReturnVoid -> 
        "return"
    | VarDef {var_symbol = Some sym} ->
        "/*VarDef*/ " ^ (string_of_symbol sym)
    | VarDef {var_name = n; var_type = t} ->
        "/*NoSymVarDef*/ " ^ n ^ ":" ^ (string_of_type t)
    | Block ss ->
        "{\n" ^ list_map string_of_stmt ss ^
        (make_ind ind) ^ "}"
    | Pair (s1, s2) ->
        "{ /* pair */\n" ^ list_map string_of_stmt [s1;s2] ^
        (make_ind ind) ^ "} /* pair */"
    | Switch (e, cc) ->
        let string_of_cc ind {cc_pattern = p; cc_stmt = s} =
          (make_ind ind) ^ "case " ^ (string_of_pattern p) ^ ":\n" ^
          (string_of_stmt (ind + 1) s) in
        "switch " ^ (string_of_exp e) ^ " {\n" ^
        list_map string_of_cc cc ^ (make_ind ind) ^ "}"
    | Let (pel, s) ->
        "let\n" ^ list_map 
          (fun ind (p, e) -> 
              (make_ind ind) ^ (string_of_pattern p) ^ " = " ^ 
              (string_of_exp e) ^ "\n") pel ^
        (make_ind ind) ^ "in {\n" ^ (string_of_stmt (ind+1) s) ^
        (make_ind ind) ^ "} /* let */\n"
    | If (e, s1, s2) ->
        "if " ^ string_of_exp e ^ "{ /* if */\n" ^ 
        (string_of_stmt (ind + 1) s1) ^
        (make_ind ind) ^ "} else {\n" ^ (string_of_stmt (ind + 1) s2) ^
        (make_ind ind) ^ "} /* if */"
    | Loop {raw_stmt = Pair ({raw_stmt = BreakIfNot c}, b)} ->
        "while " ^ (string_of_exp c) ^ " {\n" ^
        (string_of_stmt (ind+1) b) ^ 
        (make_ind ind) ^ "}"
    | Loop {raw_stmt = Pair (b, {raw_stmt = BreakIfNot c})} ->
        "do {\n" ^
        (string_of_stmt (ind+1) b) ^ 
        (make_ind ind) ^ "} while " ^ (string_of_exp c)
    | Loop b ->
        "loop {\n" ^
        (string_of_stmt (ind+1) b) ^ 
        (make_ind ind) ^ "}"
    | BreakIfNot e ->
        "if (!" ^ (string_of_exp e) ^ ") break"
    | Open m -> "open " ^ m 
  in
    (make_ind ind) ^ str ^ "\n"
and string_of_exp e =
  match e.raw_exp with
  | SimpleBinOp (o, e1, e2) ->
    let mid =
      match o with
      | IntPlus -> "+"
      | IntMinus -> "-"
      | IntMult -> "*"
      | IntDiv -> "/"
      | IntMod -> "%"
      | FloatPlus -> "+."
      | FloatMinus -> "-."
      | FloatMult -> "*."
      | FloatDiv -> "/."
      | Assign -> "="
      | Eq -> "=="
      | Ne -> "!="
      | Lt -> "<"
      | Gt -> ">"
      | Ge -> ">="
      | Le -> "<=" in
    "(" ^ (string_of_exp e1) ^ " " ^ mid ^ " " ^ (string_of_exp e2) ^ ")"
  | SimpleUnOp (o, e) ->
    let op =
      match o with
      | IntNegate -> "-"
      | FloatNegate -> "-."
    in
      "(" ^ op ^ " " ^ (string_of_exp e) ^ ")"
  | Ref {ref_symbol = Some sym} -> "/*ref*/ " ^ (string_of_symbol sym)
  | Ref {ref_name = s} -> s
  | CrossModuleRef {xmr_symbol = Some sym} ->
      "/*xmr*/ " ^ (string_of_symbol sym)
  | CrossModuleRef {xmr_symbol = None} -> assert false
  | Call (e, p) ->
      (string_of_exp e) ^ "(" ^ 
      (String.concat ", " (List.map string_of_exp p)) ^ ")"
  | IntConst s -> s
  | BoolConst(true) -> "true"
  | BoolConst(false) -> "false"
  | FloatConst s -> s
  | VoidConst -> "()"
  | StringConst s -> "\"" ^ (String.escaped s) ^ "\""
  | TupleExp el ->
      "[" ^ (String.concat ", " (List.map string_of_exp el)) ^ "]"
  | UnionExp {ur_sym = Some sym; ur_exp = e} ->
      "/*union*/ " ^ (string_of_symbol sym) ^ "["
  | UnionExp {ur_sym = None} -> assert false
  | Lambda {lambda_ret_type = r;
            lambda_parms = p;
            lambda_body = b;
            lambda_tyvars = tyv} ->
      "fun <" ^ (String.concat ", " (List.map (fun t -> "'" ^ t) tyv)) ^ 
          "> (" ^ 
        (String.concat ", " (List.map 
                      (fun p -> 
                         match p with 
                         | {parm_symbol = Some sym} -> string_of_symbol sym
                         | _ -> assert false) p)) ^ 
       ") -> " ^ (string_of_type r) ^ " is {\n" ^ 
       (string_of_stmt 1 b) ^ "}"
and string_of_pattern p =
  match p.raw_pattern with
  | VoidPattern -> "_"
  | VarPattern {vp_symbol = Some sym} -> "/*vp*/ " ^ (string_of_symbol sym)
  | VarPattern {vp_symbol = None} -> assert false
  | ExprPattern e -> string_of_exp e
  | TuplePattern {tpr_patterns = p} -> 
      "[" ^ (String.concat ", " (List.map string_of_pattern p)) ^ "]"
  | UnionPattern {upr_sym = Some sym; upr_pattern = pat} ->
      "/* up */ " ^ (string_of_symbol sym) ^ "[" ^ 
      (string_of_pattern pat) ^ "]"
  | UnionPattern {upr_sym = None} -> assert false
;;

(* More entry points for debugger. *)
let print_type f t =
  Format.pp_print_string f (string_of_type t)
let print_exp f e =
  Format.pp_print_string f (string_of_exp e)
let print_pat f p =
  Format.pp_print_string f (string_of_pattern p)
let print_stmt f s =
  Format.pp_print_string f (string_of_stmt 0 s)
let print_symbol f s =
  Format.pp_print_string f (string_of_symbol s)
;;

(* o[x/t] *)
let rec subst (x, tt) o =
  let f t =
    match t with
          TyVar y when x = y -> tt
        | TyVar _ | NamedType _ | Int | Void | Bool | Float | String 
        | AbstractType | Tuple _ | Func _ | Struct _ | OptStruct _ 
        | Union _ -> t in
  type_walk f o
;;

exception Is_FV

let is_fv v t =
  let f t =
     match t with
        TyVar x when v = x -> raise Is_FV
      | TyVar _ | Int | Void | Bool | Float | String
      | Func _ | Tuple _ | Union _ | Struct _ | OptStruct _ 
      | NamedType _ | AbstractType -> t in
  try
    ignore(type_walk f t);
    false
  with Is_FV -> true
;;

(* Unfortunetly this function cannot be recursive, because types can
 * by cyclic... *)
let rec instance_value tt =
    match tt with
        NamedType ({nt_tydef = Some tydef; 
                    nt_parms = p} as n) ->
          (try
            let mapping = List.map2 (fun x y -> (x, y)) tydef.td_tyvars p in
            List.fold_left (fun o (x, t) -> subst (x, t) o) 
                           tydef.td_type 
                           mapping
          with Invalid_argument _ ->
            err (n.nt_loc, xf "type has %d parameters, but it was instanced \
                               with %d arguments" 
                              (List.length tydef.td_tyvars) 
                              (List.length p)))
      | NamedType {nt_tydef = None} -> assert false 
      | t -> t 
;;

(* Term unification, the standard, given-from-God, algorithm :^) *)
let rec mgu (o : (id * ty) list) rrs =
  match rrs with
      r::rs ->
          (match r with
              (TyVar(x), TyVar(y)) when x = y -> 
                  mgu o rs
            | (TyVar(x), t) 
            | (t, TyVar(x)) when not (is_fv x t) ->
                  mgu ((x, t) :: (List.map (fun (xx, tt) -> 
                                                (xx, subst (x, t) tt)) o))
                                         (* Horrible error was here... *)
                      (List.map (fun (t1, t2) -> (subst (x, t) t1, 
                                                  subst (x, t) t2)) rs)
            | (Int, Int) | (Bool, Bool) | (Void, Void)
            | (Float, Float) | (String, String) -> 
                  mgu o rs
            | ((Func {ft_ret_type=r1; ft_parms=p1} as x), 
               (Func {ft_ret_type=r2; ft_parms=p2} as y)) ->
                  if List.length p1 != List.length p2 then
                    sorry (xf "cannot unify function types %s and %s \
                               with diffrent number of parameters"
                              (string_of_type x)
                              (string_of_type y))
                  else
                    mgu o ((r1, r2) :: 
                           (List.map2 (fun a b -> (a, b)) p1 p2) @ rs)
            | ((Tuple l1 as x), (Tuple l2 as y)) ->
                if List.length l1 != List.length l2 then
                  sorry (xf "cannot unfiy tuple types %s and %s of diffrent \
                             arity" 
                            (string_of_type x)
                            (string_of_type y))
                else
                  mgu o ((List.map2 (fun a b -> (a, b)) l1 l2) @ rs)
            | ((Struct(l1) as x), (Struct(l2) as y)) 
            | ((Union(l1) as x), (Union(l2) as y)) 
            | ((OptStruct(l1) as x), (OptStruct(l2) as y)) ->
                (try
                  List.iter2 (fun (t1, n1) (t2, n2) -> 
                     if n1 <> n2 then raise (Invalid_argument "cmp"))
                  l1 l2
                 with Invalid_argument _ -> 
                        sorry (xf "cannot unify %s with %s, since they \
                                   have diffrent struct/union fields"
                                   (string_of_type x)
                                   (string_of_type y)));
                  mgu o ((List.map2 (fun (t1, n1) (t2, n2) -> 
                                                    (t1, t2)) l1 l2) @ rs)
            (* First of all, try to unify at names level. *)
            | (NamedType {nt_tydef = t1; nt_parms = p1},
               NamedType {nt_tydef = t2; nt_parms = p2}) when (t1 = t2) ->
                 mgu o ((List.map2 (fun x y -> (x, y)) p1 p2) @ rs)
            (* But if it fails, try fallback to actuall types. *)
            | (((NamedType _) as x), y) 
            | (x, ((NamedType _) as y)) ->
                  mgu o ((instance_value x, instance_value y) :: rs)
            | (x, y) ->
                    sorry (xf "cannot unify %s with %s" 
                              (string_of_type x)
                              (string_of_type y)))
                                     
    | [] -> o
;;

type ty_context = {ctx_ret_type: ty;   (* Return type of function. *)
                   ctx_tyvars: id list (* Variables functions is parametrized
                                        * over. *)
                  }

(* Oh well... this function is misleading. *)
let fresh_vars t =
  let f t =
    match t with
          Func {ft_tyvars = ((_::_) as tv)} -> 
            List.fold_left (fun t x -> subst (x, 
                                              TyVar(seq_name "$")) t) 
                           t tv
        | Func _ | TyVar _ | NamedType _ | Int | Void | Bool | Float | String 
        | AbstractType | Tuple _ | Struct _ | OptStruct _ 
        | Union _ -> t in
  type_walk f t
  
(* Check whatever s and t are compatibile in given context c.
 * Return MGU. *)
let checker c s t =
  let check_if_nok lst (v, t) =
    let swaper (u, v) n =
      if n = u then v else if n = v then u else n in
    let term_swaper (u, v) t =
      match t with TyVar x -> TyVar (swaper (u, v) x) | t -> t in
    let swap (u, v) (n, t)  =
      (swaper (u, v) n, type_walk (term_swaper (u, v)) t) in
    if List.mem v c.ctx_tyvars then
      match t with
          TyVar x when not (List.mem x c.ctx_tyvars) -> 
            List.map (swap (x, v)) lst
        | _ -> sorry (xf "'%s has to be %s, while it should be \
                          assumption-free" v (string_of_type t))
    else lst in
  let m = mgu [] [(t, s)] in
  List.fold_left check_if_nok m m

(* Check if t < s. *)
let subtype c s t =
  (* I guess this is wrong (fresh_vars doesn't do exactly what it's
   * expected to...) *)
  ignore(checker c (fresh_vars t) (fresh_vars s))
  
let global_subtype s t =
  subtype {ctx_ret_type = AbstractType; ctx_tyvars = []} s t
  
let typecheck_call c f at =
  let ft = 
    match f with
        Func {ft_parms = x} -> x
      | _ -> sorry (xf "called object is not a function \
                        (it has non-functional type %s)"
                               (string_of_type f)) in
  if (List.length ft) != (List.length at) then
    sorry (xf "there is %d formal parameters and %d actual parameters in call"
              (List.length ft) (List.length at));
  let var = seq_name "$" in
  (* Hmmm... I wonder what was it suppoused to do... 
  let a = fresh_vars (Func {ft_ret_type = TyVar(var);
                            ft_parms = at;
                            ft_tyvars = [] (* Here we cheat a bit... *)
                           })  in
  let a' = match a with Func r -> Func {r with ft_tyvars = [var]} 
                      | _ -> assert false in *)
  let a' = Func {ft_ret_type = TyVar(var); ft_parms = at; ft_tyvars=[var]} in
  let m = checker c a' (fresh_vars f) in
  try
    let (_, t) = 
      (try
        List.find (fun (x, _) -> x = var) m 
       with Not_found ->
         let (n, _) = List.find (fun (_, x) -> 
                                        match x with 
                                            TyVar(y) when y = var -> true
                                          | _ -> false) m in 
         (n, TyVar n)) in
    match t with
         TyVar k -> 
           if k.[0] == '$' then
             sorry (xf "what the heck? return value of call is '%s" k)
           else
             t
       | _ -> t
  with Not_found ->
    (* hmm... shouldn't we return TyVar(var) here? *)
    sorry "return type of call is unspecified"
;;

let rec lvalue_p e =
  match e.raw_exp with
      (* Of course this might be wrong... We need some 'const' notation. *)
      Ref r -> ()
    | TupleExp t -> 
       let _ = List.map lvalue_p t  in
       ()
    | _ -> sorry "expression is not a proper lvalue"
;;

let rec strip_off ({ctx_tyvars = t} as g) l =
  match l with
      x::xs -> 
        if List.mem x t then
          strip_off g xs
        else
          x :: (strip_off g xs)
    | [] -> []
    
let stripper g t =
  let f t =
    match t with
          Func ({ft_tyvars = tt} as f) -> 
            f.ft_tyvars <- strip_off g tt;
            Func f
        | TyVar _ | NamedType _ | Int | Void | Bool | Float | String 
        | AbstractType | Tuple _  | Struct _ | OptStruct _ 
        | Union _ -> t in
  ignore(type_walk f t)

let rec typecheck_exp (g, e) =
  let subtype_p g t1 t2 =
    try
      subtype g t1 t2; true
    with Type_error _ -> false in
  try
    let (g', t) =
      match e.raw_exp with
          SimpleBinOp(op, e1, e2) ->
            let (g1, t1) = typecheck_exp (g, e1) in
            let (g2, t2) = typecheck_exp (g1, e2) in
            (match op with
                IntPlus | IntMinus | IntDiv | IntMod | IntMult ->
                  subtype g2 Int t1;
                  subtype g2 Int t2;
                  (g2, Int)
              | FloatPlus | FloatMinus | FloatDiv | FloatMult -> 
                  subtype g2 Float t1;
                  subtype g2 Float t2;
                  (g2, Float)
              | Assign ->
                  subtype g2 t1 t2;
                  lvalue_p e1;
                  (g2, t1)
              | Eq | Ne ->
                  if subtype_p g2 Bool t1 then
                    subtype g2 Bool t2
                  else if subtype_p g2 Int t1 then
                    subtype g2 Int t2
                  else if subtype_p g2 Float t1 then
                    subtype g2 Float t2
                  else
                    sorry (xf "`!=' and `==' operators supported only on \
                               bool, int and float types, not on (%s, %s)"
                              (string_of_type t1)
                              (string_of_type t2));
                  (g2, Bool)
              | Lt | Gt | Le | Ge ->
                  if subtype_p g2 Int t1 then
                    subtype g2 Int t2
                  else if subtype_p g2 Float t1 then
                    subtype g2 Float t2
                  else
                    sorry (xf "`<', '>', '<=' and `>=' operators supported \
                               only on int and float types, not on (%s, %s)"
                              (string_of_type t1)
                              (string_of_type t2));
                  (g2, Bool)
              )
        | SimpleUnOp(op, e) ->
            let (g', t) = typecheck_exp (g, e) in
            (match op with
                 IntNegate -> subtype g' Int t; (g', Int)
               | FloatNegate -> subtype g' Float t; (g', Float))
        | Ref {ref_symbol = Some sym} ->
            if sym#has_property IsUnionField then
             (* If we met sth like Nil in code (where Nil is part of some
              * union), treat it as Nil[]: *)
             (e.raw_exp <- 
                 UnionExp {ur_exp = {raw_exp = VoidConst;
                                     exp_loc = e.exp_loc;
                                     exp_type = Some Void};
                           ur_module = "";
                           ur_name = "";       (* We don't need name here, *)
                           ur_sym = Some sym}; (* as we have the symbol. *)
              typecheck_exp (g, e))
            else
             (let t = sym#get_type in
              (* Drop type variables from locals. This is becasue in 
               * 'a f( *('a)->'a x) { ... } f has type 'a->'a, where
               * 'a is some given type (it is not free). *)
              if sym#get_fnc_level > 0 then stripper g t;
              (g, t))
        | Ref {ref_symbol = None} -> assert false
        | Call(f, act) ->
            let (g', t) = typecheck_exp (g, f) in
            let (g'', tact) = list_apply typecheck_exp (g', act) in
            let ret_type = typecheck_call g'' t tact in
            (* DEBUG: *)
            (* warn(e.exp_loc, "rettype = " ^ (string_of_type ret_type)); *)
            (g'', ret_type)
        | IntConst _ -> (g, Int)
        | BoolConst _ -> (g, Bool)
        | FloatConst _ -> (g, Float)
        | StringConst _ -> (g, String)
        | VoidConst -> (g, Void)
        | CrossModuleRef {xmr_symbol = Some sym} -> (g, sym#get_type)
        | CrossModuleRef {xmr_symbol = None} -> assert false
        | TupleExp tup ->
            let rec tuple_typer (g, tu) =
              match tu with
                x::xs -> 
                  let (g', t) = typecheck_exp (g, x) in
                  let (g'', t') = tuple_typer(g', xs) in
                  (g'', t::t')
              | [] -> (g, []) in
            let (g', t) = tuple_typer (g, tup) in
            (g', Tuple t)
        | UnionExp {ur_sym = None} -> assert false
        | UnionExp ({ur_sym = Some sym} as un) -> 
            (*
              u { 'a X; 'a Y }

              switch (X[1.0]) {
              case X[1.0]: ....
              case Y[1]: // type error
              }
            *)
            if sym#has_property IsUnionField then
              let (g', t) = typecheck_exp (g, un.ur_exp) in
              let ut = fresh_vars sym#union_type.td_type in
              let (ft, _) = 
                List.find (fun (t, n) -> n = sym#get_name) 
                          (match ut with Union l -> l | _ -> assert false) in
              let a = mgu [] [(ft, t)] in
              let ut' = 
                List.fold_left (fun ut (i, t) -> subst (i, t) ut) ut a in
              (g', ut')
            else
              sorry(xf "identifier `%s' is not part of any union"
                       sym#get_name)
        | Lambda l  ->
            (* Here we need to strip off type variables, that was
             * put here by stupid parser. *)
            List.iter (fun {parm_type=t} -> stripper g t) l.lambda_parms;
            let tyvars = (strip_off g l.lambda_tyvars) @ g.ctx_tyvars in
            ignore(typecheck_stmt ({g with 
                                    ctx_ret_type = l.lambda_ret_type;
                                    ctx_tyvars = tyvars},
                                   l.lambda_body));
            (g, Func {ft_ret_type = l.lambda_ret_type;
                      ft_parms = List.map (fun {parm_type=t} -> t)
                                          l.lambda_parms;
                      ft_tyvars = tyvars}) in
    e.exp_type <- Some t;
    (g', t)
  with Type_error s -> raise (Syntax_error (e.exp_loc, s))

and typecheck_stmt (g, s) =
  try
    match s.raw_stmt with
        Skip -> 
          (g, Void)
      | Exp(e) -> 
          let (g', _) = typecheck_exp (g, e) in 
          (g', Void)
      | Return(e) -> 
        let (g', t) = typecheck_exp (g, e) in
        subtype g g.ctx_ret_type t; 
        (g', Void)
      | ReturnVoid ->
        subtype g g.ctx_ret_type Void;
        (g, Void)
      | Block(ss) ->
        let (g', _) = list_apply typecheck_stmt (g, ss) in
        (g', Void)
      | VarDef {var_type = t} ->
        (* Here we need to strip off type variables, that was
         * put here by stupid parser. *)
        stripper g t;
        (g, Void)
      | Open(_) ->
        (g, Void)
      | Pair(a, b) ->
        let (g', _) = typecheck_stmt (g, a) in
        let (g'', _) = typecheck_stmt (g', b) in
        (g'', Void)
      | BreakIfNot e ->
        let (g', t) = typecheck_exp (g, e) in
        subtype g Bool t;
        (g', Void)
      | Switch (e, pl) ->
          let (g, te) = typecheck_exp (g, e) in
          let rec typecheck_patterns (g, pl) =
            match pl with
              {cc_pattern = p; cc_stmt = s}::xs ->
                (* instance_value is called by typecheck_pattern now. *)
                let _ = typecheck_pattern(te, g, p) in
                let (g, _) = typecheck_stmt(g, s) in
                typecheck_patterns(g, xs)
            | [] -> g in
          let g = typecheck_patterns(g, pl) in
          pattern_exchaustive e.exp_loc [] pl;
          (g, Void)
      | Let (patel, st) ->
          let rec typecheck_let (g, pel) =
            match pel with
              (pat, ex) :: xs ->
                 let (g, te) = typecheck_exp (g, ex) in
                 let _ = typecheck_pattern(te, g, pat) in
                 typecheck_let (g, xs)
            | [] -> g in
          let g = typecheck_let (g, patel) in
          let (g, _) = typecheck_stmt(g, st) in
          (g, Void)
      | If(e, s1, s2) ->
          let (g1, t) = typecheck_exp (g, e) in
          let (g2, _) = typecheck_stmt (g1, s1) in
          let (g3, _) = typecheck_stmt (g2, s2) in
          subtype g Bool t;
          (g3, Void)
      | Loop s ->
        typecheck_stmt (g, s)
  with Type_error e -> raise (Syntax_error (s.stmt_loc, e))

and is_const ex =
  match ex.raw_exp with
  | SimpleBinOp(_, e1, e2) -> (is_const e1) && (is_const e2)
  | SimpleUnOp(_, e) -> is_const e
  | Lambda _  
  | Ref _ 
  | Call _ 
  | CrossModuleRef _ -> false
  | IntConst _
  | BoolConst _
  | FloatConst _
  | VoidConst 
  | StringConst _ -> true
  | TupleExp tup ->
     not (List.exists (fun e -> not (is_const e)) tup)
  | UnionExp {ur_exp = e} -> is_const e

and typecheck_pattern (tex, g, pa) =
    match pa.raw_pattern with
    | VoidPattern ->
       ()
    | VarPattern {vp_symbol = Some sym} ->
       sym#set_type tex
    | VarPattern {vp_symbol = None} -> assert false
    | ExprPattern const ->
       let (g, t) = typecheck_exp (g,const) in
       subtype g tex t;
       if is_const const then
         ()
       else
         err (const.exp_loc, "only constant expressions in pattern matching \
                              allowed")

    | TuplePattern {tpr_patterns = p} ->
        let np = List.length p in
        let loc = pa.pat_loc in
        (match (instance_value tex) with
         | Tuple t ->
             if List.length t != np then
               err(loc, xf "pattern is %d-tuple, value is %d-tuple"
                           np (List.length t))
             else
               List.iter2 (fun t p -> typecheck_pattern (t, g, p)) t p
         (* Tuple patter with empty list of patterns is special case --
          * pattern that only matches void value. So we allow it here. *)
         | Void when (np == 0) -> ()
         (* However 1-tuple isn't supported ;) (it might have occured from
          * Foo[x], but [x] there is not a pattern anymore. *)
         | _ ->
             err(loc, xf "expected tuple as value to match"))
    | UnionPattern ({upr_sym = Some sym} as un) ->
          if sym#has_property IsUnionField then
            (try
              subtype g (sym#union_type).td_type tex;
              (*debug (string_of_type tex);
                debug (string_of_type (sym#union_type).td_type);*)
              match (instance_value tex) with
              | Union l -> 
                let (t, _) = 
                         List.find (fun (t, n) -> sym#get_name = n) l in
                typecheck_pattern (t, g, un.upr_pattern)
              | _ -> assert false
             with Type_error s -> err(pa.pat_loc, s))
          else
            err(pa.pat_loc, xf "identifier `%s' is not part of any union"
                               un.upr_name)
    | UnionPattern {upr_sym = None} -> assert false 

and pattern_exchaustive loc ran_l plist =
  match plist with
  | {cc_pattern = pat}::xs ->
      let pat_r = range_of_pattern pat in
      let (cmp, nran_l) = join_lists ran_l [pat_r] in
      (if (cmp <= 0) then 
        let lo = pat.pat_loc in
        warn (lo, "pattern clause is unused"));
      pattern_exchaustive loc nran_l xs
  | [] ->
      let hd = 
        (match ran_l with
        | x::_ -> x
        | [] -> assert false
        ) in
      if (hd <> AlphaRange) then
        let (_, example) = generate_example 0 ran_l in
        warn (loc, xf "this pattern matching isn't exchaustive, here is \
                       example of pattern which isn't matched: %s" example)
      else ()

and range_of_pattern pat =
  match pat.raw_pattern with
  | VoidPattern ->
      AlphaRange
  | VarPattern _ ->
      AlphaRange
  | ExprPattern const ->
      (match const.raw_exp with
      | IntConst _ -> IntRange
      | BoolConst b -> BoolRange b
      | FloatConst _ -> FloatRange
      | StringConst _ -> StringRange
      | _ -> AlphaRange
      )
  | TuplePattern {tpr_patterns = p} ->
      let lister x = [x] in
      let rl = List.map range_of_pattern p in
      let rll = List.map lister rl in
      TupleRange rll
  | UnionPattern {upr_sym = Some sym; upr_pattern = un_pat} ->
      let rp = range_of_pattern un_pat in
      UnionRange ([sym], [rp]) 
  | UnionPattern {upr_sym = None} -> assert false 

and generate_example eq rpats =
  let rpat = 
    match rpats with
    | x::xs -> x
    | [] -> AlphaRange in
  match rpat with
  | AlphaRange -> (eq, "_")
  | IntRange -> (1, "some int")
  | FloatRange -> (1, "some float")
  | StringRange -> (1, "some string")
  | BoolRange b -> 
      if (b && eq == 1 || not b && eq == 0) then (1, "true")
      else (1, "false")
  | UnionRange (syml, [AlphaRange]) ->
      let rec find_absent_name l1 l2 =
        match l2 with
        | x::xs ->
            let exist = List.exists (function y -> (x = y#get_name)) l1 in
            if exist then find_absent_name l1 xs
            else x
        | [] -> assert false in
     let ty = (List.hd syml)#union_type.td_type in
     let tsym_l = 
       match ty with
       | Union l -> List.map (function (_, x) -> x) l
       | _ -> assert false in
     if (eq == 0) then
       let nm_name = find_absent_name syml tsym_l in
       let fu y =
         match y with
         | UnionRange ([x],_) when (x#get_name = nm_name) -> true
         | _ -> false in
       let absent_pat = List.filter fu rpats in
       if (absent_pat = []) then (1, nm_name ^ " _")
       else
         generate_example 0 absent_pat
     else 
       (1, (string_of_symbol (List.hd syml)) ^ "_")
  | UnionRange ([sym], rplist) ->
      let (neq, str) = generate_example eq rplist in
      (neq, (string_of_symbol sym) ^ " " ^ str)
  | UnionRange (_, _) -> assert false
  | TupleRange rtl ->
      let rec stringer eql li =
        match li with
        | x::xs -> 
            let (eq', str) = generate_example eql x in
            let (eq'', strr) = stringer eq' xs in
            (eq'', str ^ ", " ^ strr)
        | [] -> (eql, "") in
      let (neq, st) = stringer eq rtl in
      (neq, "[" ^ st ^ "]")

(*
  old < new -> 1
  old = new -> 0
  old > new -> -1
  joined(r1, r2) -> 2
*)
and fold_tuple t =
  let fold_list l =
     let l = List.map fold_tuple l in
     match l with
     | [x] -> [x]
     | [] -> []
     | x::xs -> let (_, j) = join_lists [x] xs in j in
  match t with
  | TupleRange l ->
     let l = List.map fold_list l in
     if List.exists (fun x -> x <> [AlphaRange]) l then
       TupleRange l
     else
       AlphaRange
  | UnionRange (syms, pats) ->
      UnionRange (syms, fold_list pats)
  | AlphaRange | IntRange | StringRange
  | FloatRange | BoolRange _ -> t
  
and join_ranges oldr newr =
  match (oldr, newr) with
  | (AlphaRange, AlphaRange) -> (0, AlphaRange)
  | (AlphaRange, _) -> (-1, AlphaRange)
  | (_, AlphaRange) -> (1, AlphaRange)
  | (UnionRange (old_s, [AlphaRange]), UnionRange (new_s, [AlphaRange])) ->
     let rec uniq l1 l2 =
       match l2 with
       | x::xs -> 
         if List.memq x l1 then 
           uniq l1 xs 
         else
           uniq (x :: l1) xs
       | [] -> l1 in
     let joined = List.sort compare (uniq old_s new_s) in
     let ty = (List.hd old_s)#union_type.td_type in
     let needed_length = 
       match ty with
       | Union l -> List.length l
       | _ -> assert false in
     let joined_length = List.length joined in
     let l1_length = List.length old_s in
     let l2_length = List.length new_s in
     let res =
        if l1_length == joined_length && 
           l2_length == joined_length then 
             0
        else if l1_length == joined_length then
             -1
        else if l2_length == joined_length then
             1
        else
             2 in
     if joined_length == needed_length then
       (res, AlphaRange)
     else
       (res, UnionRange (joined, [AlphaRange]))
  | (UnionRange (old_s, [AlphaRange]), UnionRange ([s], _)) ->
      if List.memq s old_s then
        (-1, oldr)
      else
        raise Cannot_join
  | (UnionRange ([s], _), UnionRange (new_s, [AlphaRange])) ->
      if List.memq s new_s then
        (1, newr)
      else
        raise Cannot_join
  | (UnionRange ([old_s], old_l), UnionRange ([new_s], new_l)) -> 
      if old_s == new_s then
        let (res, l) = join_lists old_l new_l in
        (res, UnionRange ([new_s], l))
      else
        raise Cannot_join
  | (UnionRange _, UnionRange _) -> assert false
  | (TupleRange old_l, TupleRange new_l) ->
      let rec join state ret l =
        match l with
        | (o, n)::xs ->
          let (cmp, p) = join_lists o n in
          (match (cmp, state) with
           | (0, _) -> 
               join state (p :: ret) xs
           | (1, 0)
           | (1, 1)
           | (-1, 0) 
           | (-1, -1) -> 
               join cmp (p :: ret) xs
           | (2, 0) ->
               join cmp (p :: ret) xs
           | _ -> raise Cannot_join)
        | [] -> (state, fold_tuple (TupleRange (List.rev ret))) in
      join 0 [] (List.combine old_l new_l)
  | (BoolRange b1, BoolRange b2) ->
    if (b1 == b2) then (0, BoolRange b1)
    else (2, AlphaRange)
  | (IntRange, _) -> raise Cannot_join
  | (FloatRange, _) -> raise Cannot_join
  | (StringRange, _) -> raise Cannot_join
  | (_, _) -> assert false
    
and join_lists old_l new_l =
  let res = List.fold_left append_to_list old_l new_l in
  let res = List.sort compare res in
  if (res = old_l) && (res = new_l) then
    (0, res)
  else if (res = old_l) then
    (-1, res)
  else if (res = new_l) then
    (1, res)
  else
    (2, res)
  
and append_to_list old_l new_r =
  let rec append_rec res old new_r =
    match old with
    | x::xs ->
        (try
           let (_, p) = join_ranges x new_r in
           append_to_list (res @ xs) p
        with Cannot_join ->
          append_rec (x :: res) xs new_r)
    | [] -> new_r :: res in
  append_rec [] old_l (fold_tuple new_r)
;;

(* Introduce new, artificial, type variables in type definition. This helps
 * up catching errors. ('$v* should not appear in type of any expression). *)
let td_fresh_variables td =
  let map : (string, string) Hashtbl.t = Hashtbl.create 16 in
  let f t =
    match t with
    | TyVar a ->
      if not (Hashtbl.mem map a) then
        Hashtbl.add map a (seq_name "$v"); 
      TyVar (Hashtbl.find map a)
    | t -> t in
  td.td_type <- type_walk f td.td_type;
  td.td_tyvars <- 
    List.map (fun n -> 
                try 
                  Hashtbl.find map n 
                with Not_found ->
                  err(td.td_loc, xf "unused type variable '%s" n)) 
             td.td_tyvars
;;                                                    

(* chini_* (check if initialized) functions check whatever all variables
 * are properly initialized before they are used. They are somewhat 
 * simplistic (or conservative). This could be improved I guess. 
 *
 * `return' stmt is treated as initialization of *all* variables.
 * We maintain list of initialized variables, and walk with it down the
 * expression/stmt terms. return stmt introduces Everything node to this
 * list, it should be present in functions returning non-void, after 
 * precessing top-level stmt (each path ends with return then).
 *
 * Precessing `break' stmts in loops are somewhat tricky. They push 
 * special Breaker node with level of loop nesting to the initialized
 * variables list. Then, variables initialized in loop, are these
 * before first Breaker pushed (they are after last breaker in the list),
 * with Breaker prepended if it would break upper level loop (we don't
 * use it right now, as there is no break stmt in Gont right now..., 
 * however for and while loops are translated to loop(if (c) break; ...))
 *
 * Handling of Lambda expression is currently somewhat too conservative.
 * For example:
 *
 *      int f() 
 *      { 
 *              int m;
 *              void g() 
 *              { 
 *                      return m;       // gontc complains here
 *              }
 *              m = 5;
 *              return g();
 *      }
 *
 * I don't know how could we fix it, though...
 *)

type ini =
    Symbol of symbol
  | Breaker of int
  | Everything

type chini_ctx = {cc_initialized: ini list;
                  cc_this_stmt_initialized: ini list;
                  cc_loop_level: int}

(* Check initialization in sequence of stmts/exprs (depending on chini_one
 * functional parameter). Seqeunce is given in lst. *)
let chini_list chini_one gctx lst =
  let rec chini_list_tail ini ctx lst =
    match lst with
    | x::xs ->
      let ctx = chini_one ctx x in
      let ini = ctx.cc_this_stmt_initialized @ ini in
      let ctx = {ctx with cc_this_stmt_initialized = [];
                          cc_initialized = ini @ gctx.cc_initialized} in
      chini_list_tail ini ctx xs
    | [] -> ini in
  let ini = chini_list_tail [] gctx lst in
  {gctx with cc_this_stmt_initialized = ini}
  
(* Check initialization in one stmt. *)
let rec chini_stmt ctx s =
    (* Find minimum of v1 v2, treat -1 as +Infty *)
    let strange_min v1 v2 =
      if v1 == -1 then v2
      else 
        if v2 == -1 then v1
        else
          if v1 < v2 then v1
          else v2 in
    let normalize lst =
      let rec normalize_tail min res lst =
        match lst with
        | (Breaker k) :: xs -> 
          let min = strange_min min k in
          normalize_tail min xs xs
        | x :: xs ->
          normalize_tail min res xs
        | [] ->
          (min, res) in
      normalize_tail (-1) lst lst in
    (* Find intersection of two given, sorted lists. *)
    let intersection lst1 lst2 =
      let rec find_is res lst1 lst2 =
        match lst1 with
        | x::xs ->
          (match lst2 with
           | y::ys ->
             let c = compare x y in
             if c < 0 then
               find_is res xs lst2
             else
               if c > 0 then
                 find_is res lst1 ys
               else
                 find_is (x::res) xs ys
           | [] -> res)
        | [] -> res in
      if List.mem Everything lst1 then 
         lst2
      else 
         if List.mem Everything lst2 then
           lst1
         else
           List.rev (find_is [] lst1 lst2) in
    (* Check initialization in mutually exclusive branches given in l. *)
    let chini_branches gctx l =
      let rec chini_branches_tail (min, ini) l =
        match l with
        | x::xs ->
          let ctx = chini_stmt gctx x in
          let (min', ini') = normalize ctx.cc_this_stmt_initialized in
          let ini = intersection ini (List.sort compare ini') in
          let min = strange_min min min' in
           chini_branches_tail (min, ini) xs
        | [] -> 
          let ini =
            if min == -1 then ini
            else (Breaker min) :: ini in
          {gctx with cc_this_stmt_initialized = ini} in
       chini_branches_tail (-1, [Everything]) l in
    match s.raw_stmt with
    | Skip 
    | VarDef _ 
    | Open _ -> ctx
    | Exp e -> chini_exp ctx e
    | Return e -> 
        let ctx = chini_exp ctx e in
        let ini = Everything :: ctx.cc_this_stmt_initialized in
        {ctx with cc_this_stmt_initialized = ini}
    | ReturnVoid ->
        {ctx with cc_this_stmt_initialized = [Everything]}
    | Block ss -> 
        chini_list chini_stmt ctx ss
    | Pair (a, b) ->
        chini_list chini_stmt ctx [a; b]
    | Loop s ->
        let ctx_for_s = {ctx with cc_loop_level = ctx.cc_loop_level + 1} in
        let ctx' = chini_stmt ctx_for_s s in
        let (min, ini) = normalize ctx'.cc_this_stmt_initialized in
        let ini =
          if min == -1 or min == ctx.cc_loop_level + 1 then 
            ini 
          else 
            (Breaker min) :: ini in
        {ctx with cc_this_stmt_initialized = ini}
    | BreakIfNot e ->
        let ctx = chini_exp ctx e in
        let ini = (Breaker ctx.cc_loop_level) :: 
                  ctx.cc_this_stmt_initialized in
        {ctx with cc_this_stmt_initialized = ini}
    | Switch (e, pl) ->
        let gctx = chini_exp ctx e in
        (* Fortunetly we don't have any expressions, involving variables 
         * in patterns (right now...), so focus on stmts only. *)
        chini_branches gctx (List.map (fun {cc_stmt = s} -> s) pl)
    | Let (pl, s) ->
        let ctx = chini_list chini_exp ctx (List.map (fun (p, e) -> e) pl) in
        chini_stmt ctx s
    | If(e, s1, s2) ->
        let ctx = chini_exp ctx e in
        chini_branches ctx [s1; s2]
        
(* Check initialization in one expr. *)
and chini_exp ctx exp =
  match exp.raw_exp with
  (* Fscking special case... otherwise local function cannot be recursive. *)
  | SimpleBinOp (Assign, 
                 {raw_exp = Ref {ref_symbol = Some sym}}, 
                 ({raw_exp = Lambda _} as e2)) ->
      let ctx' = {ctx with cc_initialized =
                                (Symbol sym) :: ctx.cc_initialized} in
      let ini = (chini_exp ctx' e2).cc_this_stmt_initialized in
      {ctx with cc_this_stmt_initialized = (Symbol sym) :: ini}
  | SimpleBinOp (op, e1, e2) ->
      (* We might want to change this, if we define order of evaluation
       * of binary operators arguments... *)
      let ini = (chini_exp ctx e2).cc_this_stmt_initialized in
      let ini =
        match (op, e1.raw_exp) with
          | (Assign, Ref {ref_symbol = Some sym}) -> (Symbol sym) :: ini
          | _ -> (chini_exp ctx e1).cc_this_stmt_initialized @ ini in
      {ctx with cc_this_stmt_initialized = ini}
  | UnionExp {ur_exp = e}
  | SimpleUnOp (_, e) ->
      (* We don't care about the actuall operator, union symbol or 
       * whatever... *)
      chini_exp ctx e
  | Ref {ref_symbol = Some sym} ->
      if sym#has_property IsUnionField or 
         sym#has_property IsInitialized or 
         sym#get_fnc_level == 0 then
        ctx
      else
        if List.mem (Symbol sym) ctx.cc_initialized or
           List.mem Everything ctx.cc_initialized then
             ctx
        else
             err (exp.exp_loc, xf "variable %s might be used uninitialized"
                                  sym#get_name)
  | Ref {ref_symbol = None} -> assert false
  | Call (f, act) ->
      chini_list chini_exp ctx (f :: act)
  | TupleExp tup ->
      chini_list chini_exp ctx tup
  | Lambda {lambda_body = s} ->
      (* FIXME: check if we properly `return' from Lambda. *)
      ignore(chini_stmt ctx s);
      ctx
  | IntConst _
  | BoolConst _
  | FloatConst _
  | StringConst _
  | VoidConst
  | CrossModuleRef _ -> ctx
;;

(* Check initialization of variables in function. *)
let chini_func ret_type body =
  let ctx = chini_stmt {cc_initialized = []; 
                        cc_this_stmt_initialized = [];
                        cc_loop_level = 1} body in
  match ret_type with
  | Void -> ()
  | _ ->
    if List.mem Everything ctx.cc_this_stmt_initialized then
      ()
    else
      err (body.stmt_loc, "this function can finish without return")
;;

let typecheck_decl d =
  match d with
      FuncDecl f -> 
        ignore(typecheck_stmt ({ctx_ret_type = f.fn_ret_type; 
                                ctx_tyvars = f.fn_tyvars}, f.fn_body));
        chini_func f.fn_ret_type f.fn_body
    | GlobalDecl _ -> ()
    | ExternC _ -> ()
    | FuncProto _ -> ()
    | TopLevelOpen _ -> ()
    | TypeDef td -> td_fresh_variables td
;;

let type_of e =
  match e.exp_type with
      Some t -> t
    | None -> assert false
;;
